from typing import List, Union, Tuple, Generator
import random
from collections import namedtuple

RangeArgs = Union[Tuple[int, int], Tuple[str, str]]
BracketMatch = namedtuple("BracketMatch", ["start", "stop", "level"])
CharMatch = namedtuple("CharMatch", ["index", "level"])

class MultiRange():

    def __init__(self, ranges:List[range]=[]):
        self.ranges = self._clean(ranges)

    def append(self, r):
        self.ranges = self._clean(self.ranges+[r])

    def _clean(self, ranges) -> List[range]:
        '''sort ranges and remove duplicates'''
        ranges = list(ranges)
        ranges.sort(key=lambda r: r.start)

        cleaned = []
        for r in ranges:
            if len(cleaned) == 0:
                cleaned.append(r)
            else:
                last_r = cleaned[-1]
                if r.stop <= last_r.stop:
                    continue
                elif r.start >= last_r.stop:
                    cleaned.append(r)
                elif r.start < last_r.stop:
                    cleaned.append(range(last_r.stop, r.stop))
        
        return cleaned

    def random_choice(self) -> int:
        i = random.randint(0, len(self)-1)
        return self[i]

    def __getitem__(self, index:int):
        for r in self.ranges:
            if index < len(r):
                return r[index]
            else:
                index -= len(r)
        raise IndexError()

    def __len__(self) -> int:
        length = 0
        for r in self.ranges:
            length += len(r)
        return length

    def inverted(self, start:int, stop:int) -> 'MultiRange':
        new_ranges = []
        for r in self.ranges:
            if start >= r.start:
                start = r.stop
            elif start < r.start and r.start < stop:
                new_ranges.append(range(start, r.start))
                start = r.stop
            elif start < r.start and r.start > stop:
                new_ranges.append(range(start, stop))
            
            if start >= stop:
                break

        if start < stop:
            new_ranges.append(range(start, stop)) 
            
        return MultiRange(new_ranges)

def multirange(*range_args:List[RangeArgs]) -> MultiRange:
    ranges = []
    for [start, stop] in range_args:
        if type(start) == str:
            start = ord(start)
        if type(stop) == str:
            stop = ord(stop)+1
        ranges.append(range(start, stop))
    return MultiRange(ranges)

def char_range(start:str, stop:str) -> range:
    return range(ord(start), ord(stop)+1)


def find_brackets(s, ob='(', cb=')') -> Generator[BracketMatch, None, None]:
    stack = []

    for i, c in enumerate(s):
        if c == ob:
            stack.append(i)
        elif c == cb:
            yield BracketMatch(stack.pop(), i, len(stack))

    if stack:
        raise IndexError('too many opening brackets')

def find_char(s:str, search:str, level_up="([", level_down="])"):
    stack = []
    for i, c in enumerate(s):
        if c in level_up:
            stack.append(i)
        elif c in level_down:
            stack.pop()
        elif c == search:
            yield CharMatch(index=i, level=len(stack))





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = 'genrex',
    packages = ['genrex'],
    license='MIT',
    description = 'Genrex generates matching strings to a given regular expressions.',
    author = 'abichinger',
    author_email = 'abichinger@example.com',
    url = 'https://github.com/abichinger/genrex',
    long_description=long_description,
    long_description_content_type="text/markdown",
    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.3',
)


import random
from typing import List, Union
from genrex.util import multirange, MultiRange, char_range, find_brackets, find_char
from enum import Enum
import sys

class PatternBase():
    
    def random(self) -> str:
        index = random.randint(0, len(self)-1)
        return self[index]

    def __getitem__(self, index:int) -> str:
        pass

    def __len__(self) -> int:
        pass
        
class Char(PatternBase): 
    
    def __init__(self, char:int):
        self.char = chr(char)

    def random(self) -> str:
        return self.char

    def __getitem__(self, index:int) -> str:
        if index != 0:
            raise IndexError
        return self.char

    def __len__(self) -> int:
        return 1

class Set():

    def __init__(self, multi_range:Union[range, MultiRange]):
        self.multi_range = multi_range

    def __getitem__(self, index:int) -> str:
        return chr(self.multi_range[index])

    def __len__(self) -> int:
        return len(self.multi_range)

    def inverted(self, start:int, stop:int) -> int:
        return Set(self.multi_range.inverted(start, stop))

    def append(self, r):
        self.multi_range.append(r)
    
class RepeatingPattern(PatternBase):
    
    def __init__(self, pattern:PatternBase, m:int, n:int):
        self.p = pattern
        self.m = m
        self.n = n

    def __getitem__(self, index:int) -> str:
        if index > len(self)-1:
            raise IndexError
 
        for m in range(self.m, self.n+1):
            l = len(self.p)**m
            if l > index:
                break
            else:
                index -= l

        item = ''
        for m in range(0, m):
            item += self.p[index%len(self.p)]
            index = int(index/len(self.p))

        return item

    def __len__(self) -> int:
        cardinality = 0
        for i in range(self.m, self.n+1):
            cardinality += len(self.p)**i
        return cardinality

class Pattern(PatternBase):

    def __init__(self, children:List[PatternBase]):
        self.children = children

    def append(self, child):
        self.children.append(child)

    def __getitem__(self, index:int) -> str:
        if index > len(self)-1:
            raise IndexError

        item = ''
        for child in self.children:
            item += child[index%len(child)]
            index = int(index/len(child))
        return item

    def __len__(self) -> int:
        cardinality = 1
        for child in self.children:
            child_len = len(child)
            if child_len > 0:
                cardinality *= child_len
        
        return cardinality

class Either(PatternBase):

    def __init__(self, a:PatternBase, b:PatternBase):
        self.a = a
        self.b = b
    
    def __getitem__(self, index:int) -> str:
        if index < len(self.a):
            return self.a[index]
        else:
            return self.b[index-len(self.a)]

    def __len__(self) -> int:
        return len(self.a) + len(self.b)


import sre_parse
from sre_constants import *

def parse(pattern:str, flags=0, max_depth=15) -> PatternBase:
    '''parses a regular expression using the built in sre engine
    
    :param pattern: re-style regular expression
    :param flags: re flags
    :param max_depth: limit of repeating expressions
    '''

    p = sre_parse.parse(pattern)
    return _parse(p, flags, max_depth)

def _parse(p:sre_parse.SubPattern, flags, max_depth):

    subp = Pattern([])

    for op, av in p.data:
        if op is LITERAL:
            subp.append(Char(av))
        if op is IN:
            char_set = Set(MultiRange())
            for op, a in av:
                if op is RANGE:
                    char_set.append(range(a[0], a[1]+1))
                if op is LITERAL:
                    char_set.append(range(a, a+1))
            subp.append(char_set)
        if op is SUBPATTERN:
            subp.append(_parse(av[-1], flags, max_depth))
        if op is BRANCH:
            branches = av[-1]
            a = _parse(branches[0], flags, max_depth)
            for b in branches[1:]:
                a = Either(a, _parse(b, flags, max_depth))
            subp.append(a)
        if op is MAX_REPEAT or op is MIN_REPEAT:
            a = _parse(av[-1], flags, max_depth)
            if av[1] is MAXREPEAT:
                subp.append(RepeatingPattern(a, av[0], max_depth))
            else:
                subp.append(RepeatingPattern(a, av[0], av[1]))
        if op is ANY:
            new_line = Set(MultiRange([range(ord('\n'), ord('\n')+1)]))
            subp.append(new_line.inverted(0, sys.maxunicode))

    return subp